{-- Haskell98 --}

-- Number-parameterized vectors: arbitrary-precision decimal numbers, 
-- type-constructor-based encoding (the size is the phantom type)

module ArbPrecDecT
	(D0(..), D1(..), D2(..), D3(..), D4(..), D5(..), D6(..), 
	 D7(..), D8(..), D9(..), Sz(..), 
	 Digits(..), Card(..), CardC, cardc,
	 Vec,          -- but not the data constructor!	 
	 vec, listVec, -- builders
	 vlength_t, vlength, vat, vupdate, velems, 
	 vzipWith, 
	 v12c, v12i, v13i -- test vectors
	 ) where

import Data.Array

-- Types and constructors for individual digits
-- Depending on the context, D5 stands either for the type for numeral 5,
-- or for the data constructor for numeral 5

data D0 a = D0 a
data D1 a = D1 a
data D2 a = D2 a
data D3 a = D3 a
data D4 a = D4 a
data D5 a = D5 a
data D6 a = D6 a
data D7 a = D7 a
data D8 a = D8 a
data D9 a = D9 a

class Digits ds where		-- class of digit sequences
    ds2num:: (Num a) => ds -> a -> a     -- CPS style
   
data Sz = Sz			-- zero size (or the Nil of the sequence)

instance Digits Sz where
    ds2num _ acc = acc
    
instance (Digits ds) => Digits (D0 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc)
instance (Digits ds) => Digits (D1 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 1)
instance (Digits ds) => Digits (D2 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 2)
instance (Digits ds) => Digits (D3 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 3)
instance (Digits ds) => Digits (D4 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 4)
instance (Digits ds) => Digits (D5 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 5)
instance (Digits ds) => Digits (D6 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 6)
instance (Digits ds) => Digits (D7 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 7)
instance (Digits ds) => Digits (D8 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 8)
instance (Digits ds) => Digits (D9 ds) where
    ds2num dds acc = ds2num (t22 dds) (10*acc + 9)
 
t22::(f x)   -> x; t22 = undefined

-- Class of non-negative numbers
-- This is a restriction on Digits. It is not possible to make
-- such a restriction in SML.
class {- (Digits c) => -} Card c where
    c2num:: (Num a) => c -> a
    
instance Card Sz where c2num c = ds2num c 0
--instance (NonZeroDigit d,Digits (d ds)) => Card (Sz (d ds)) where
instance (Digits ds) => Card (D1 ds) where c2num c = ds2num c 0
instance (Digits ds) => Card (D2 ds) where c2num c = ds2num c 0
instance (Digits ds) => Card (D3 ds) where c2num c = ds2num c 0
instance (Digits ds) => Card (D4 ds) where c2num c = ds2num c 0
instance (Digits ds) => Card (D5 ds) where c2num c = ds2num c 0
instance (Digits ds) => Card (D6 ds) where c2num c = ds2num c 0
instance (Digits ds) => Card (D7 ds) where c2num c = ds2num c 0
instance (Digits ds) => Card (D8 ds) where c2num c = ds2num c 0
instance (Digits ds) => Card (D9 ds) where c2num c = ds2num c 0

-- Support for "generic" cards
-- We introduce a data constructor CardC_unused merely for the sake of
-- Haskell98. With the GHC extension, we can simply omit the data
-- constructor and keep the type CardC purely abstract and phantom.
data CardC c1 c2 = CardC_unused

cardc:: (Card c1, Card c2) => c1 -> c2 -> CardC c1 c2
cardc = undefined
cardc1::CardC c1 c2 -> c1; cardc1 = undefined
cardc2::CardC c1 c2 -> c2; cardc2 = undefined

instance (Card c1, Card c2) => Card (CardC c1 c2) where
    c2num c12 = c2num (cardc1 c12) + c2num (cardc2 c12)


-- The following is just as before.

-- The data constructor Vec is NOT to be exported
newtype Vec size a = Vec (Array Int a) deriving Show

vec:: (Card size) => size -> a -> Vec size a
vec size elem = Vec $ listArray (0,(c2num size)-1) $ repeat elem

-- The following is an internal function. It does not check
-- that the list `elems' has the `size' elements.
-- The function is used internally when we have proven that |elems| is 
-- indeed size or when it's OK to truncate elems to size
listVec':: (Card size) => size -> [a] -> Vec size a
listVec' size elems = Vec $ listArray (0,(c2num size)-1) elems

-- The following is an exported function. 
listVec:: (Card size) => size -> [a] -> Vec size a
listVec size elems | not (c2num size == length elems) =
		       error "listVec: static/dynamic sizes mismatch"
listVec size elems = listVec' size elems

-- gives static, compile-time size
vlength_t::(Card size) => Vec size a -> size
vlength_t _ = undefined

-- gives run-time size
vlength:: Vec size a -> Int
vlength (Vec arr) = let (0,last) = bounds arr in last+1

vat::  Vec size a -> Int -> a
vat (Vec arr) i = arr ! i

velems:: Vec size a -> [a]
velems (Vec v) = elems v

vupdate:: Vec size a -> [(Int,a)] -> Vec size a
vupdate (Vec arr) ul = Vec $ arr // ul

-- The type of vzipWith assures the safety side-condition of listVec'
vzipWith:: (Card size) => 
   (a->b->c) -> (Vec size a) -> (Vec size b) -> (Vec size c)
vzipWith f va vb = 
   listVec' (vlength_t va) $ zipWith f (velems va) (velems vb)

-- Tests

sample_v = listVec (D2 Sz) [True,False]

v12c = listVec (D1 $ D2 Sz) $ take 12 ['a'..'z']
v12i = listVec (D1 $ D2 Sz) [1..12]
v13i = listVec (D1 $ D3 Sz) [1..13]

-- :t v12c

test1 = vlength v12c; test1' = vlength_t v12c
test2 = [vat v12c 0, vat v12c 1, vat v12c 2]

test3 = vzipWith (\c i -> (toEnum $ fromEnum c + fromIntegral i)::Char)
	         v12c v12i
test4 = vzipWith (+) v12i v12i

-- The following leads to a type error, as expected:
-- test5 = vzipWith (+) v12i v13i

{--
*ArbPrecDecT> test4
Vec (array (0,11) [(0,2),(1,4),(2,6),(3,8),(4,10),(5,12),(6,14),(7,16),(8,18),(9,20),(10,22),(11,24)])
*ArbPrecDecT> :type test4
Vec (D1 (D2 Sz)) Int
--}
