{-- Haskell98 --}

-- Number-parameterized vectors: unary numbers, type-constructor-based
-- encoding (the size is the phantom type)

module UnaryT (Zero(..), Succ(..), Card(..),
	       Vec,			-- Vec constructor is not exported!
	       listVec,
	       vec,
	       vlength_t, vlength,
	       velems, vat, vzipWith,
	       vc, vs, (&+)
	       )
	       where

import Data.Array

data Zero = Zero
data Succ a = Succ a

class Card c where
    c2num:: (Num a) => c -> a -- convert to a number

cpred::(Succ c) -> c
cpred = undefined

instance Card Zero where 
    c2num _ = 0
instance (Card c) => Card (Succ c) where
    c2num x = 1 + c2num (cpred x)
    

-- The data constructor Vec is NOT exported
newtype Vec size a = Vec (Array Int a) deriving Show

-- The following is an internal function. It does not check
-- that the list `elems' has the `size' elements.
-- The function is used internally when we have proven that |elems| is 
-- indeed size or when it's OK to truncate elems to size
listVec':: (Card size) => size -> [a] -> Vec size a
listVec' size elems = Vec $ listArray (0,(c2num size)-1) elems

-- The following is an exported function. 
listVec:: (Card size) => size -> [a] -> Vec size a
listVec size elems | not (c2num size == length elems) =
		       error "listVec: static/dynamic sizes mismatch"
listVec size elems = listVec' size elems


vec:: (Card size) => size -> a -> Vec size a
vec size elem = listVec' size $ repeat elem

-- gives static, compile-time size
vlength_t:: Vec size a -> size
vlength_t _ = undefined

-- gives run-time size
vlength:: Vec size a -> Int
vlength (Vec arr) = let (0,last) = bounds arr in last+1

velems:: Vec size a -> [a]
velems (Vec v) = elems v

vat::  Vec size a -> Int -> a
vat (Vec arr) i = arr ! i

-- The type of vzipWith assures the safety side-condition of listVec'
vzipWith:: Card size => 
	   (a->b->c) -> Vec size a -> Vec size b -> Vec size c
vzipWith f va vb = 
   listVec' (vlength_t va) $ zipWith f (velems va) (velems vb)

sample_v = listVec (Succ (Succ Zero)) [True,False]

-- More convenient notation

infixl 3 &+
data VC size a = VC size [a]

vs = VC Zero []
(&+) (VC size lst) x = VC (Succ size) (x:lst)
vc (VC size lst) = listVec' size (reverse lst)

v3c = vc $ vs &+ 'a' &+ 'b' &+ 'c'
v3i = vc $ vs &+ 1 &+ 2 &+ 3
v4i = vc $ vs &+ 1 &+ 2 &+ 3 &+ 4

-- :t v3c
-- Note from the inferred type the separation of the skeleton and the content

test1 = vlength v3c; test1' = vlength_t v3c
test2 = [vat v3c 0, vat v3c 1, vat v3c 2]

test3 = vzipWith (\c i -> (toEnum $ fromEnum c + fromIntegral i)::Char) v3c v3i
test4 = vzipWith (+) v3i v3i

-- The following leads to a type error, as expected:
-- test5 = vzipWith (+) v3i v4i

